# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABCMeta, abstractmethod
from typing import Union

_PROTOCOLS = {}
_DEFAULT_PROTOCOL = None
_DEFAULT_PROTOCOL_NAME = "serobj"

CUSTOM_NONE_TYPE = type("NoneType", (), {})()


def _set_default_protocol(protocol):
    global _DEFAULT_PROTOCOL
    _DEFAULT_PROTOCOL = protocol


def load(proto_obj, raise_exception=True):
    try:
        protocol_name = proto_obj["meta"]["protocol"]
        protocol_version = proto_obj["meta"]["version"]
    except (KeyError, TypeError):
        if raise_exception:
            raise ValueError(
                "object `proto_obj` has not protocol and/or version information"
            )
        else:
            return None

    try:
        protocol = _PROTOCOLS[protocol_name][protocol_version]
    except KeyError:
        if raise_exception:
            raise ValueError("`proto_obj` protocol not found")
        else:
            return None

    if protocol.is_compatible_proto_obj(proto_obj, raise_exception=raise_exception):
        return protocol(proto_obj).object

    return None


def dump(obj, protocol=None):
    protocol = protocol or _DEFAULT_PROTOCOL

    return protocol(obj).proto_object


class SerObjProtocolMeta(ABCMeta):
    def __new__(mcs, name, bases, dct):
        protocol_name = dct.get("PROTOCOL_NAME", None) or _DEFAULT_PROTOCOL_NAME
        protocol_version = dct.get("PROTOCOL_VERSION", None)

        dct["PROTOCOL_NAME"] = protocol_name
        dct["PROTOCOL_VERSION"] = protocol_version or 0

        cls = type.__new__(mcs, name, bases, dct)

        if protocol_version is not None:
            if protocol_name in _PROTOCOLS and protocol_version in _PROTOCOLS.get(
                protocol_name, {}
            ):
                raise RuntimeError(
                    "Protocol '{} version {}' already registered".format(
                        protocol_name, protocol_version
                    )
                )

            proto_versions = _PROTOCOLS.get(protocol_name, {})
            proto_versions[protocol_version] = cls
            _PROTOCOLS[protocol_name] = proto_versions

            if (
                not _DEFAULT_PROTOCOL
                or protocol_name == _DEFAULT_PROTOCOL_NAME
                and _DEFAULT_PROTOCOL.PROTOCOL_VERSION < protocol_version
            ):
                _set_default_protocol(cls)

        return cls


class SerObjProtocolBase(metaclass=SerObjProtocolMeta):
    PROTOCOL_NAME = None
    PROTOCOL_VERSION = None

    def __init__(self, obj):
        if self.is_compatible_proto_obj(obj, raise_exception=False):
            self._meta = obj["meta"]
            self._payload = obj["payload"]
            self._obj = self._construct(obj["payload"])
        else:
            self._meta = {
                "protocol": self.PROTOCOL_NAME,
                "version": self.PROTOCOL_VERSION,
            }
            self._payload = self._deconstruct(obj)
            self._obj = obj

    @classmethod
    @abstractmethod
    def is_compatible_proto_obj(
        cls, proto_obj: dict, raise_exception: bool = True
    ) -> bool:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _deconstruct(cls, obj: object) -> dict:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _construct(cls, proto_obj_payload: dict) -> object:
        raise NotImplementedError

    @property
    def object(self) -> object:
        return self._obj

    @property
    def proto_object(self) -> Union[dict, None]:
        if not self._payload or self._payload is CUSTOM_NONE_TYPE:
            return None
        return {"meta": self._meta, "payload": self._payload}
