# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
from collections import Iterable, Iterator, Mapping
from types import BuiltinFunctionType, FunctionType, LambdaType, MethodType
from typing import Callable, Union

from cerberus import Validator
from serobj.attrs import SerobjAttrs
from serobj.protocol.base import CUSTOM_NONE_TYPE, SerObjProtocolBase
from serobj.utils.path_to_object import get_object_source_info, import_object_source
from six import binary_type, integer_types, string_types

_VALIDATION_SCHEMA = {
    "source": {
        "type": "dict",
        "schema": {
            "name": {"type": "string", "required": True},
            "module": {"type": "string"},
        },
    },
    "representation": {
        "type": "dict",
        "required": True,
        "schema": {
            "args": {
                "type": "list",
                "schema": {
                    "type": "dict",
                    "schema": {
                        "name": {"type": "string", "required": True},
                        "value": {"nullable": True},
                    },
                },
            },
            "attrs": {"type": "dict", "schema": {"keyschema": {"type": "string"}}},
        },
    },
}
_SCHEMA_VALIDATOR = Validator(_VALIDATION_SCHEMA, allow_unknown=True)
_RECURSION_TOKEN = "__SEROBJ__RECURSION__"
_INIT_LIKE_NEW_TOKEN = "__SEROBJ__INIT_LIKE_NEW__"


class ZeroSerObjProtocol(SerObjProtocolBase):
    PROTOCOL_VERSION = 0

    @classmethod
    def _get_obj_main_attrs(cls, obj):
        def _filter(x):
            x_obj = getattr(obj, x)
            return (
                not x.startswith("_")
                and not x_obj.__class__.__module__  # exclude protected and private
                == "builtins"
                and not x_obj.__class__.__name__.startswith(  # exclude builtin attrs
                    "_"
                )
                and not hasattr(  # exclude attrs with hidden classes  # exclude descriptors (like @property)
                    getattr(obj.__class__, x, object), "__get__"
                )
            )

        return {k: getattr(obj, k) for k in filter(_filter, dir(obj))}

    @classmethod
    def _get_reduce_params(cls, obj, param_no, default=None):
        reduced_fn = getattr(obj, "__reduce__", None)
        if reduced_fn:
            try:
                reduced = reduced_fn()
                if reduced and reduced[0] != obj.__class__:
                    reduced = ()
                if not isinstance(reduced, Iterable) or isinstance(
                    reduced, string_types
                ):
                    reduced = ()
            except TypeError:
                reduced = ()
            return reduced[param_no] if len(reduced) > param_no else default
        else:
            return default

    @classmethod
    def _get_attrs_filter(cls, obj: object) -> Union[FunctionType, LambdaType, bool]:
        attrs_filter_key, attrs_filter = SerobjAttrs.ATTRS_FILTER_KEY.search_attr_in(
            obj, default=None
        )
        if attrs_filter is None:
            return True
        if isinstance(attrs_filter, Mapping):

            def map_filter(_, attr_name, attr_val):
                return attrs_filter[attr_name] == attr_val

            attrs_filter = map_filter
        elif isinstance(attrs_filter, Iterable):

            def iter_filter(_, attr_name, __):
                return attr_name in attrs_filter

            attrs_filter = iter_filter
        elif not callable(attrs_filter):
            raise TypeError(
                "'{}' should be Callable or Mapping or Iterable".format(
                    attrs_filter_key
                )
            )
        return attrs_filter

    @classmethod
    def _get_attrs_by_str(cls, obj: object, key_string: str):
        if key_string == "from_reduce":
            attrs = cls._get_reduce_params(obj, 2, {})
        elif key_string == "main":
            attrs = cls._get_obj_main_attrs(obj)
        else:
            attrs = getattr(obj.__class__, key_string, None) or getattr(obj, key_string)

        return attrs

    @classmethod
    def _get_default_attrs(cls, obj: object):
        attrs = cls._get_reduce_params(obj, 2, {})
        if not attrs or not isinstance(attrs, dict):
            attrs = getattr(obj, "__dict__", None)
        return attrs

    @classmethod
    def _get_attrs(cls, obj: object):
        attrs_filter = cls._get_attrs_filter(obj)
        attrs_key, attrs = SerobjAttrs.ATTRS_KEY.search_attr_in(obj, default=None)

        if isinstance(attrs, str):
            attrs = cls._get_attrs_by_str(obj, attrs)
        if callable(attrs):
            attrs = attrs(obj)
        if attrs is None:
            attrs = cls._get_default_attrs(obj)
            if not isinstance(attrs, dict):
                raise TypeError("bad __reduce__ value")

        if isinstance(attrs, Mapping):
            attrs = dict(attrs)
        elif isinstance(attrs, Iterable):
            attrs = {k: getattr(obj, k) for k in attrs}

        if not isinstance(attrs, dict):
            raise TypeError(
                "'{}' should be: str or None or Callable or Mapping or Iterable".format(
                    attrs_key
                )
            )

        return {
            k: v
            for k, v in attrs.items()
            if attrs_filter is True
            or callable(attrs_filter)
            and attrs_filter(obj, k, v)
        }

    @classmethod
    def _get_args_by_str(cls, obj: object, key_string: str):
        if key_string == "from_reduce":
            args = cls._get_reduce_params(obj, 1, [])
        elif key_string == "default":
            args = None
        else:
            args = getattr(obj.__class__, key_string, None) or getattr(obj, key_string)

        return args

    @classmethod
    def _get_default_args(cls, obj: object, args_type: str):
        args = cls._get_reduce_params(obj, 1, None)

        if args and not isinstance(args, Iterable):
            raise TypeError("bad __reduce__ value")

        # else from __getnewargs__ or from __getinitargs__
        if args is None:
            args_method_name = "__get{}args__".format(args_type)
            getargs_fn = getattr(obj, args_method_name, None)
            if getargs_fn:
                args = getargs_fn()

            if args and not isinstance(args, Iterable):
                raise TypeError("bad {} value".format(args_method_name))

        return args

    @classmethod
    def _get_args(cls, obj: object, args_enum_key: SerobjAttrs):
        assert args_enum_key in (
            SerobjAttrs.NEW_ARGS_KEY,
            SerobjAttrs.INIT_ARGS_KEY,
        ), "'args_key' not in (SerobjAttrs.NEW_ARGS_KEY, SerobjAttrs.INIT_ARGS_KEY)"

        args_key, args = args_enum_key.search_attr_in(obj, default=CUSTOM_NONE_TYPE)

        if args == CUSTOM_NONE_TYPE:
            if args_enum_key == SerobjAttrs.NEW_ARGS_KEY:
                args = cls._get_default_args(obj, "new")
            else:
                args = cls._get_default_args(obj, "init")

        if isinstance(args, str):
            args = cls._get_args_by_str(obj, args)

        if callable(args):
            if getattr(args, "__self__", CUSTOM_NONE_TYPE) == obj:
                args = args()
            else:
                args = args(obj)

        if isinstance(args, Mapping):
            args = [{"name": k, "value": v} for k, v in dict(args).items()]
        elif isinstance(args, Iterable):
            args = list(args)
            if not all(
                map(
                    lambda x: isinstance(x, dict) and "name" in x and "value" in x, args
                )
            ):
                args = [{"name": "", "value": v} for v in args]

        if args is not None and not isinstance(args, list):
            raise TypeError(
                "'{}' should be: str or None or Callable or Mapping or Iterable".format(
                    args_key
                )
            )

        return args

    @classmethod
    def _get_init_args(cls, obj):
        return cls._get_args(obj, SerobjAttrs.INIT_ARGS_KEY)

    @classmethod
    def _get_new_args(cls, obj):
        return cls._get_args(obj, SerobjAttrs.NEW_ARGS_KEY)

    @classmethod
    def _get_representation(cls, obj: object, context: dict):
        if obj is None:
            return None

        if isinstance(obj, (Iterable, Mapping)):
            return {"attrs": {}, "init_args": [{"name": "", "value": obj}]}

        repr_obj = {"attrs": {}}

        attrs = cls._get_attrs(obj)
        if attrs:
            for k, v in attrs.items():
                if id(v) in context["processed"]:
                    prepared_attr_val = _RECURSION_TOKEN
                else:
                    prepared_attr_val = cls._deconstruct(v, context)
                    if prepared_attr_val is CUSTOM_NONE_TYPE:
                        continue
                repr_obj["attrs"][k] = prepared_attr_val

        if isinstance(obj, type):
            return repr_obj

        new_args = cls._get_new_args(obj)
        init_args = cls._get_init_args(obj)

        def _process_args(args):
            _processed_args = []
            for arg in args:
                prepared_arg_val = cls._deconstruct(arg["value"], context)
                if prepared_arg_val is CUSTOM_NONE_TYPE:
                    raise ValueError(
                        "'{}' is not serializable by serobj [{}.{} protocol]".format(
                            arg["value"], cls.PROTOCOL_NAME, cls.PROTOCOL_VERSION
                        )
                    )

                arg["value"] = prepared_arg_val

                _processed_args.append(arg)
            return _processed_args

        if new_args is not None:
            repr_obj["new_args"] = _process_args(new_args)

        if init_args is not None:
            if init_args == new_args:
                repr_obj["init_args"] = _INIT_LIKE_NEW_TOKEN
            else:
                repr_obj["init_args"] = _process_args(init_args)

        if new_args is None and init_args is None:
            repr_obj["new_args"] = []

        return repr_obj

    @classmethod
    def _deconstruct(cls, obj, context=None):
        if isinstance(
            obj,
            (
                string_types
                + (binary_type,)
                + integer_types
                + (float, bool, None.__class__)
            ),
        ):
            return obj

        context = dict(context or {})
        context.update({"processed": context.get("processed", []) + [id(obj)]})

        source_obj, repr_obj = obj, obj

        if isinstance(obj, Mapping):
            repr_obj = {
                cls._deconstruct(k, context): cls._deconstruct(v, context)
                for k, v in obj.items()
            }
        elif isinstance(obj, Iterator):
            source_obj, repr_obj = iter([]), [cls._deconstruct(v, context) for v in obj]
        elif isinstance(obj, Iterable):
            repr_obj = [cls._deconstruct(v, context) for v in obj]
        elif obj.__class__ in [FunctionType, MethodType, BuiltinFunctionType]:
            repr_obj = None
        elif isinstance(obj, type):
            pass
        elif obj.__class__.__name__.startswith("_"):
            return CUSTOM_NONE_TYPE

        if isinstance(obj, (list, dict)):
            return repr_obj

        return {
            "source": get_object_source_info(source_obj),
            "representation": cls._get_representation(repr_obj, context),
        }

    @classmethod
    def _construct_no_proto(cls, obj):
        if isinstance(
            obj,
            (
                string_types
                + (binary_type,)
                + integer_types
                + (float, bool, None.__class__)
            ),
        ):
            return obj

        if isinstance(obj, Mapping):
            return {cls._construct(k): cls._construct(v) for k, v in obj.items()}
        elif isinstance(obj, Iterable):
            return [cls._construct(v) for v in obj]

        return obj

    @classmethod
    def _construct(cls, obj):
        if not isinstance(obj, dict):
            return cls._construct_no_proto(obj)

        if "source" in obj and "representation" in obj:
            source = obj["source"]
            representation = obj["representation"]
        else:
            return cls._construct_no_proto(obj)

        obj_source = import_object_source(source["full_path"])

        if not representation:
            return obj_source

        attrs = representation.get("attrs", {})
        new_args = representation.get("new_args", None)
        init_args = representation.get("init_args", None)

        def _construct_args(args):
            call_args, call_kwargs = [], {}
            for arg in args:
                arg_name, arg_value = arg["name"], cls._construct(arg["value"])
                if arg_name:
                    call_kwargs[arg_name] = arg_value
                else:
                    call_args.append(arg_value)
            return call_args, call_kwargs

        def _get_method(enum_key, default):
            method_key, method = enum_key.search_attr_in(obj_source, default=default)
            if isinstance(method, str):
                method = getattr(obj_source, method)
            if not isinstance(method, Callable):
                raise TypeError("'{}' should be str or Callable".format(method_key))
            return method

        if new_args is not None or init_args is not None:

            method_new = _get_method(SerobjAttrs.CONSTRUCT_NEW_FN, obj_source.__new__)
            method_init = _get_method(
                SerobjAttrs.CONSTRUCT_INIT_FN, obj_source.__init__
            )

            if new_args is None:
                if init_args == _INIT_LIKE_NEW_TOKEN:
                    obj = obj_source
                else:
                    i_args, i_kwargs = _construct_args(init_args)
                    obj = obj_source(*i_args, **i_kwargs)
            elif init_args is None:
                n_args, n_kwargs = _construct_args(new_args)
                obj = method_new(obj_source, *n_args, **n_kwargs)
            else:
                n_args, n_kwargs = _construct_args(new_args)

                if init_args == _INIT_LIKE_NEW_TOKEN:
                    i_args, i_kwargs = n_args, n_kwargs
                else:
                    i_args, i_kwargs = _construct_args(init_args)

                obj = method_new(obj_source, *n_args, **n_kwargs)
                method_init(obj, *i_args, **i_kwargs)
        else:
            obj = obj_source

        for attr_name, attr_value in attrs.items():
            if attr_value == _RECURSION_TOKEN:
                setattr(obj, attr_name, obj)
            else:
                constructed_attr = cls._construct(attr_value)
                setattr(obj, attr_name, constructed_attr)

        return obj

    @classmethod
    def is_compatible_proto_obj(
        cls, proto_obj: dict, raise_exception: bool = True
    ) -> bool:
        if (
            isinstance(proto_obj, dict)
            and "meta" in proto_obj
            and "payload" in proto_obj
        ):
            if _SCHEMA_VALIDATOR.validate(proto_obj["payload"] or {}):
                return True
            if raise_exception:
                raise ValueError(
                    "SCHEMA_VALIDATOR: {}".format(json.dumps(_SCHEMA_VALIDATOR.errors))
                )
        else:
            if raise_exception:
                raise ValueError(
                    "`proto_obj` argument is invalid 'Serialized Proto Object'"
                )
        return False
